@@ -35,6 +35,41 @@
 
 #define min(x, y) ((x) < (y) ? (x) : (y))
 
+int tag__is_base_type(const struct tag *tag, const struct cu *cu)
+{
+	switch (tag->tag) {
+	case DW_TAG_base_type:
+		return 1;
+
+	case DW_TAG_typedef: {
+		const struct tag *type = cu__type(cu, tag->type);
+
+		if (type == NULL)
+			return 0;
+		return tag__is_base_type(type, cu);
+	}
+	}
+	return 0;
+}
+
+bool tag__is_array(const struct tag *tag, const struct cu *cu)
+{
+	switch (tag->tag) {
+	case DW_TAG_array_type:
+		return true;
+
+	case DW_TAG_const_type:
+	case DW_TAG_typedef: {
+		const struct tag *type = cu__type(cu, tag->type);
+
+		if (type == NULL)
+			return 0;
+		return tag__is_array(type, cu);
+	}
+	}
+	return 0;
+}
+
 const char *cu__string(const struct cu *cu, strings_t s)
 {
 	if (cu->dfops && cu->dfops->strings__ptr)
@@ -289,6 +324,15 @@ void namespace__delete(struct namespace *space, struct cu *cu)
 	tag__delete(&space->tag, cu);
 }
 
+void __type__init(struct type *type)
+{
+	INIT_LIST_HEAD(&type->node);
+	INIT_LIST_HEAD(&type->type_enum);
+	type->sizeof_member = NULL;
+	type->member_prefix = NULL;
+	type->member_prefix_len = 0;
+}
+
 struct class_member *
 	type__find_first_biggest_size_base_type_member(struct type *type,
 						       const struct cu *cu)
@@ -331,7 +375,7 @@ reevaluate:
 		case DW_TAG_rvalue_reference_type:
 		case DW_TAG_volatile_type: {
 			struct tag *tag = cu__type(cu, type->type);
-			if (type == NULL) {
+			if (tag == NULL) {
 				tag__id_not_found_fprintf(stderr, type->type);
 				continue;
 			}
@@ -420,12 +464,17 @@ static int ptr_table__add_with_id(struct ptr_table *pt, void *ptr,
 		if (entries == NULL)
 			return -ENOMEM;
 
+		/* Zero out the new range */
+		memset(entries + pt->allocated_entries * sizeof(void *), 0,
+		       (allocated_entries - pt->allocated_entries) * sizeof(void *));
+
 		pt->allocated_entries = allocated_entries;
 		pt->entries = entries;
 	}
 
 	pt->entries[id] = ptr;
-	++pt->nr_entries;
+	if (id >= pt->nr_entries)
+		pt->nr_entries = id + 1;
 	return 0;
 }
 
@@ -694,6 +743,30 @@ struct tag *cu__find_enumeration_by_sname_and_size(const struct cu *cu,
 	return NULL;
 }
 
+struct tag *cu__find_enumeration_by_name(const struct cu *cu, const char *name, type_id_t *idp)
+{
+	uint32_t id;
+	struct tag *pos;
+
+	if (name == NULL)
+		return NULL;
+
+	cu__for_each_type(cu, id, pos) {
+		if (pos->tag == DW_TAG_enumeration_type) {
+			const struct type *type = tag__type(pos);
+			const char *tname = type__name(type, cu);
+
+			if (tname && strcmp(tname, name) == 0) {
+				if (idp != NULL)
+					*idp = id;
+				return pos;
+			}
+		}
+	}
+
+	return NULL;
+}
+
 struct tag *cu__find_struct_by_sname(const struct cu *cu, strings_t sname,
 				     const int include_decls, type_id_t *idp)
 {
@@ -727,6 +800,54 @@ found:
 
 }
 
+struct tag *cu__find_type_by_name(const struct cu *cu, const char *name, const int include_decls, type_id_t *idp)
+{
+	if (cu == NULL || name == NULL)
+		return NULL;
+
+	uint32_t id;
+	struct tag *pos;
+	cu__for_each_type(cu, id, pos) {
+		struct type *type;
+
+		if (!tag__is_type(pos))
+			continue;
+
+		type = tag__type(pos);
+		const char *tname = type__name(type, cu);
+		if (tname && strcmp(tname, name) == 0) {
+			if (!type->declaration)
+				goto found;
+
+			if (include_decls)
+				goto found;
+		}
+	}
+
+	return NULL;
+found:
+	if (idp != NULL)
+		*idp = id;
+	return pos;
+}
+
+struct tag *cus__find_type_by_name(const struct cus *cus, struct cu **cu, const char *name,
+				   const int include_decls, type_id_t *id)
+{
+	struct cu *pos;
+
+	list_for_each_entry(pos, &cus->cus, node) {
+		struct tag *tag = cu__find_type_by_name(pos, name, include_decls, id);
+		if (tag != NULL) {
+			if (cu != NULL)
+				*cu = pos;
+			return tag;
+		}
+	}
+
+	return NULL;
+}
+
 static struct tag *__cu__find_struct_by_name(const struct cu *cu, const char *name,
 					     const int include_decls, bool unions, type_id_t *idp)
 {
@@ -886,6 +1007,8 @@ size_t tag__size(const struct tag *tag, const struct cu *cu)
 	size_t size;
 
 	switch (tag->tag) {
+	case DW_TAG_string_type:
+		return tag__string_type(tag)->nr_entries;
 	case DW_TAG_member: {
 		struct class_member *member = tag__class_member(tag);
 		if (member->is_static)
@@ -1119,6 +1242,13 @@ int ftype__has_parm_of_type(const struct ftype *ftype, const type_id_t target,
 {
 	struct parameter *pos;
 
+	if (ftype->tag.tag == DW_TAG_subprogram) {
+		struct function *func = (struct function *)ftype;
+
+		if (func->btf)
+			ftype = tag__ftype(cu__type(cu, ftype->tag.type));
+	}
+
 	ftype__for_each_parameter(ftype, pos) {
 		struct tag *type = cu__type(cu, pos->tag.type);
 
@@ -1297,6 +1427,9 @@ static size_t tag__natural_alignment(struct tag *tag, const struct cu *cu)
 {
 	size_t natural_alignment = 1;
 
+	if (tag == NULL) // Maybe its a non supported type, like DW_TAG_subrange_type, ADA stuff
+		return natural_alignment;
+
 	if (tag__is_pointer(tag)) {
 		natural_alignment = cu->addr_size;
 	} else if (tag->tag == DW_TAG_base_type) {
@@ -1307,7 +1440,8 @@ static size_t tag__natural_alignment(struct tag *tag, const struct cu *cu)
 		natural_alignment = type__natural_alignment(tag__type(tag), cu);
 	} else if (tag->tag == DW_TAG_array_type) {
 		tag = tag__strip_typedefs_and_modifiers(tag, cu);
-		natural_alignment = tag__natural_alignment(tag, cu);
+		if (tag != NULL) // Maybe its a non supported type, like DW_TAG_subrange_type, ADA stuff
+			natural_alignment = tag__natural_alignment(tag, cu);
 	}
 
 	/*
@@ -1335,8 +1469,13 @@ static size_t type__natural_alignment(struct type *type, const struct cu *cu)
 		if (member->tag.tag == DW_TAG_inheritance &&
 		    member->virtuality == DW_VIRTUALITY_virtual)
 			continue;
+		if (member->is_static) continue;
 
 		struct tag *member_type = tag__strip_typedefs_and_modifiers(&member->tag, cu);
+
+		if (member_type == NULL) // Maybe its a DW_TAG_subrange_type, ADA stuff still not supported
+			continue;
+
 		size_t member_natural_alignment = tag__natural_alignment(member_type, cu);
 
 		if (type->natural_alignment < member_natural_alignment)
@@ -1359,6 +1498,13 @@ void type__check_structs_at_unnatural_alignments(struct type *type, const struct
 	type__for_each_member(type, member) {
 		struct tag *member_type = tag__strip_typedefs_and_modifiers(&member->tag, cu);
 
+		if (member_type == NULL) {
+			// just be conservative and ignore
+			// Found first when a FORTRAN95 DWARF file was processed
+			// and the DW_TAG_string_type wasn't yet supported
+			continue;
+		}
+
 		if (!tag__is_struct(member_type))
 			continue;
 
@@ -1377,7 +1523,7 @@ void type__check_structs_at_unnatural_alignments(struct type *type, const struct
 bool class__infer_packed_attributes(struct class *cls, const struct cu *cu)
 {
 	struct type *ctype = &cls->type;
-	struct class_member *pos, *last = NULL;
+	struct class_member *pos;
 	uint16_t max_natural_alignment = 1;
 
 	if (!tag__is_struct(class__tag(cls)))
@@ -1499,6 +1645,70 @@ struct class_member *type__find_member_by_name(const struct type *type,
 	return NULL;
 }
 
+static int strcommon(const char *a, const char *b)
+{
+	int i = 0;
+
+	while (*a != '\0' && *a == *b) {
+		++a;
+		++b;
+		++i;
+	}
+
+	return i;
+}
+
+void enumeration__calc_prefix(struct type *enumeration, const struct cu *cu)
+{
+	if (enumeration->member_prefix)
+		return;
+
+	const char *previous_name = NULL, *curr_name = "";
+	int common_part = INT32_MAX;
+	struct enumerator *entry;
+
+	type__for_each_enumerator(enumeration, entry) {
+		const char *curr_name = enumerator__name(entry, cu);
+
+		if (previous_name) {
+			int curr_common_part = strcommon(curr_name, previous_name);
+			if (common_part > curr_common_part)
+				common_part = curr_common_part;
+
+		}
+
+		previous_name = curr_name;
+	}
+
+	enumeration->member_prefix = strndup(curr_name, common_part);
+	enumeration->member_prefix_len = common_part == INT32_MAX ? 0 : common_part;
+}
+
+void enumerations__calc_prefix(struct list_head *enumerations)
+{
+	struct tag_cu_node *pos;
+
+	list_for_each_entry(pos, enumerations, node)
+		enumeration__calc_prefix(tag__type(pos->tc.tag), pos->tc.cu);
+}
+
+const char *enumeration__prefix(struct type *enumeration, const struct cu *cu)
+{
+	if (!enumeration->member_prefix)
+		enumeration__calc_prefix(enumeration, cu);
+
+	return enumeration->member_prefix;
+}
+
+uint16_t enumeration__prefix_len(struct type *enumeration, const struct cu *cu)
+{
+	if (!enumeration->member_prefix)
+		enumeration__calc_prefix(enumeration, cu);
+
+	return enumeration->member_prefix_len;
+}
+
+
 uint32_t type__nr_members_of_type(const struct type *type, const type_id_t type_id)
 {
 	struct class_member *pos;
@@ -1561,6 +1771,9 @@ static int list__for_all_tags(struct list_head *list, struct cu *cu,
 {
 	struct tag *pos, *n;
 
+	if (list_empty(list) || !list->next)
+		return 0;
+
 	list_for_each_entry_safe_reverse(pos, n, list, node) {
 		if (tag__has_namespace(pos)) {
 			struct namespace *space = tag__namespace(pos);
@@ -1647,8 +1860,8 @@ int cus__load_dir(struct cus *cus, struct conf_load *conf,
 		    strcmp(entry->d_name, "..") == 0)
 		    continue;
 
-		snprintf(pathname, sizeof(pathname), "%s/%s",
-			 dirname, entry->d_name);
+		snprintf(pathname, sizeof(pathname), "%.*s/%s",
+			 (int)(sizeof(pathname) - sizeof(entry->d_name) - 1), dirname, entry->d_name);
 
 		err = lstat(pathname, &st);
 		if (err != 0)
@@ -1683,8 +1896,8 @@ extern struct debug_fmt_ops dwarf__ops, ctf__ops, btf_elf__ops;
 
 static struct debug_fmt_ops *debug_fmt_table[] = {
 	&dwarf__ops,
-	&ctf__ops,
 	&btf_elf__ops,
+	&ctf__ops,
 	NULL,
 };
 
@@ -1740,7 +1953,7 @@ int cus__load_file(struct cus *cus, struct conf_load *conf,
 	}
 
 	while (debug_fmt_table[i] != NULL) {
-		if (conf->conf_fprintf)
+		if (conf && conf->conf_fprintf)
 			conf->conf_fprintf->has_alignment_info = debug_fmt_table[i]->has_alignment_info;
 		if (debug_fmt_table[i]->load_file(cus, conf, filename) == 0)
 			return 0;
@@ -1985,25 +2198,6 @@ static int filename__sprintf_build_id(const char *pathname, char *sbuild_id)
 	return build_id__sprintf(build_id, sizeof(build_id), sbuild_id);
 }
 
-/* asnprintf consolidates asprintf and snprintf */
-static int asnprintf(char **strp, size_t size, const char *fmt, ...)
-{
-	va_list ap;
-	int ret;
-
-	if (!strp)
-		return -EINVAL;
-
-	va_start(ap, fmt);
-	if (*strp)
-		ret = vsnprintf(*strp, size, fmt, ap);
-	else
-		ret = vasprintf(strp, fmt, ap);
-	va_end(ap);
-
-	return ret;
-}
-
 #define zfree(ptr) ({ free(*ptr); *ptr = NULL; })
 
 static int vmlinux_path__nr_entries;
@@ -2080,6 +2274,19 @@ static int cus__load_running_kernel(struct cus *cus, struct conf_load *conf)
 	int i, err = 0;
 	char running_sbuild_id[SBUILD_ID_SIZE];
 
+	if ((!conf || conf->format_path == NULL || strncmp(conf->format_path, "btf", 3) == 0) &&
+	    access("/sys/kernel/btf/vmlinux", R_OK) == 0) {
+		int loader = debugging_formats__loader("btf");
+		if (loader == -1)
+			goto try_elf;
+
+		if (conf && conf->conf_fprintf)
+			conf->conf_fprintf->has_alignment_info = debug_fmt_table[loader]->has_alignment_info;
+
+		if (debug_fmt_table[loader]->load_file(cus, conf, "/sys/kernel/btf/vmlinux") == 0)
+			return 0;
+	}
+try_elf:
 	elf_version(EV_CURRENT);
 	vmlinux_path__init();
 
