@@ -13,6 +13,7 @@
 #include <obstack.h>
 #include <dwarf.h>
 #include <elfutils/libdwfl.h>
+#include <sys/types.h>
 
 #include "dutil.h"
 #include "list.h"
@@ -54,6 +55,12 @@ struct conf_load {
 
 /** struct conf_fprintf - hints to the __fprintf routines
  *
+ * @count - Just like 'dd', stop pretty printing input after 'count' records
+ * @skip - Just like 'dd', skip 'count' records when pretty printing input
+ * @seek_bytes - Number of bytes to seek, if stdin only from start, when we have --pretty FILE, then from the end as well with negative numbers,
+ * 		 may be of the form $header.MEMBER_NAME when using with --header.
+ * @size_bytes - Number of bytes to read, similar to seek_bytes, and when both are in place, first seek seek_bytes then read size_bytes
+ * @range - data structure field in --header to determine --seek_bytes and --size_bytes, must have 'offset' and 'size' fields
  * @flat_arrays - a->foo[10][2] becomes a->foo[20]
  * @classes_as_structs - class f becomes struct f, CTF doesn't have a "class"
  * @cachelinep - pointer to current cacheline, so that when expanding types we keep track of it,
@@ -68,7 +75,13 @@ struct conf_fprintf {
 	int32_t	   type_spacing;
 	int32_t	   name_spacing;
 	uint32_t   base_offset;
+	uint32_t   count;
 	uint32_t   *cachelinep;
+	const char *seek_bytes;
+	const char *size_bytes;
+	const char *header_type;
+	const char *range;
+	uint32_t   skip;
 	uint8_t	   indent;
 	uint8_t	   expand_types:1;
 	uint8_t	   expand_pointers:1;
@@ -120,6 +133,9 @@ struct tag *cus__find_struct_by_name(const struct cus *cus, struct cu **cu,
 				     type_id_t *id);
 struct tag *cus__find_struct_or_union_by_name(const struct cus *cus, struct cu **cu,
 					      const char *name, const int include_decls, type_id_t *id);
+struct tag *cu__find_type_by_name(const struct cu *cu, const char *name, const int include_decls, type_id_t *idp);
+struct tag *cus__find_type_by_name(const struct cus *cus, struct cu **cu, const char *name,
+				   const int include_decls, type_id_t *id);
 struct function *cus__find_function_at_addr(const struct cus *cus,
 					    uint64_t addr, struct cu **cu);
 void cus__for_each_cu(struct cus *cus, int (*iterator)(struct cu *cu, void *cookie),
@@ -340,6 +356,7 @@ struct tag *cu__find_base_type_by_sname_and_size(const struct cu *cu,
 						 strings_t name,
 						 uint16_t bit_size,
 						 type_id_t *idp);
+struct tag *cu__find_enumeration_by_name(const struct cu *cu, const char *name, type_id_t *idp);
 struct tag *cu__find_enumeration_by_sname_and_size(const struct cu *cu,
 						   strings_t sname,
 						   uint16_t bit_size,
@@ -379,6 +396,12 @@ struct tag {
 	void		 *priv;
 };
 
+// To use with things like type->type_enum == perf_event_type+perf_user_event_type
+struct tag_cu {
+	struct tag	 *tag;
+	struct cu	 *cu;
+};
+
 void tag__delete(struct tag *tag, struct cu *cu);
 
 static inline int tag__is_enumeration(const struct tag *tag)
@@ -479,6 +502,7 @@ static inline int tag__is_tag_type(const struct tag *tag)
 {
 	return tag__is_type(tag) ||
 	       tag->tag == DW_TAG_array_type ||
+	       tag->tag == DW_TAG_string_type ||
 	       tag->tag == DW_TAG_base_type ||
 	       tag->tag == DW_TAG_const_type ||
 	       tag->tag == DW_TAG_pointer_type ||
@@ -656,6 +680,7 @@ struct variable {
 	enum vscope	 scope;
 	struct location	 location;
 	struct hlist_node tool_hnode;
+	struct variable  *spec;
 };
 
 static inline struct variable *tag__variable(const struct tag *tag)
@@ -763,7 +788,7 @@ void ftype__delete(struct ftype *ftype, struct cu *cu);
 void ftype__add_parameter(struct ftype *ftype, struct parameter *parm);
 size_t ftype__fprintf(const struct ftype *ftype, const struct cu *cu,
 		      const char *name, const int inlined,
-		      const int is_pointer, const int type_spacing,
+		      const int is_pointer, const int type_spacing, bool is_prototype,
 		      const struct conf_fprintf *conf, FILE *fp);
 size_t ftype__fprintf_parms(const struct ftype *ftype,
 			    const struct cu *cu, int indent,
@@ -784,6 +809,8 @@ struct function {
 	uint8_t		 external:1;
 	uint8_t		 accessibility:2; /* DW_ACCESS_{public,protected,private} */
 	uint8_t		 virtuality:2; /* DW_VIRTUALITY_{none,virtual,pure_virtual} */
+	uint8_t		 declaration:1;
+	uint8_t		 btf:1;
 	int32_t		 vtable_entry;
 	struct list_head vtable_node;
 	/* fields used by tools */
@@ -816,8 +843,8 @@ static __pure inline int tag__is_function(const struct tag *tag)
  * @func: struct function instance to iterate
  * @pos: struct parameter iterator
  */
-#define function__for_each_parameter(func, pos) \
-	ftype__for_each_parameter(&func->proto, pos)
+#define function__for_each_parameter(func, cu, pos) \
+	ftype__for_each_parameter(func->btf ? tag__ftype(cu__type(cu, func->proto.tag.type)) : &func->proto, pos)
 
 const char *function__name(struct function *func, const struct cu *cu);
 
@@ -909,6 +936,16 @@ static __pure inline int tag__is_class_member(const struct tag *tag)
 	return tag->tag == DW_TAG_member;
 }
 
+int tag__is_base_type(const struct tag *tag, const struct cu *cu);
+bool tag__is_array(const struct tag *tag, const struct cu *cu);
+
+struct class_member_filter;
+
+struct tag_cu_node {
+	struct list_head node;
+	struct tag_cu	 tc;
+};
+
 /**
  * struct type - base type for enumerations, structs and unions
  *
@@ -917,6 +954,12 @@ static __pure inline int tag__is_class_member(const struct tag *tag)
  * @nr_tags: number of tags
  * @alignment: DW_AT_alignement, zero if not present, gcc emits since circa 7.3.1
  * @natural_alignment: For inferring __packed__, normally the widest scalar in it, recursively
+ * @sizeof_member: Use this to find the size of the record
+ * @type_member: Use this to select a member from where to get an id on an enum to find a type
+ * 		 to cast for, needs to be used with the upcoming type_enum.
+ * @type_enum: enumeration(s) to use together with type_member to find a type to cast
+ * @member_prefix: the common prefix for all members, say in an enum, this should be calculated on demand
+ * @member_prefix_len: the lenght of the common prefix for all members
  */
 struct type {
 	struct namespace namespace;
@@ -926,6 +969,12 @@ struct type {
 	uint16_t	 nr_static_members;
 	uint16_t	 nr_members;
 	uint32_t	 alignment;
+	struct class_member *sizeof_member;
+	struct class_member *type_member;
+	struct class_member_filter *filter;
+	struct list_head type_enum;
+	char 		 *member_prefix;
+	uint16_t	 member_prefix_len;
 	uint16_t	 natural_alignment;
 	bool		 packed_attributes_inferred;
 	uint8_t		 declaration; /* only one bit used */
@@ -934,6 +983,8 @@ struct type {
 	uint8_t		 resized:1;
 };
 
+void __type__init(struct type *type);
+
 static inline struct class *type__class(const struct type *type)
 {
 	return (struct class *)type;
@@ -1045,6 +1096,12 @@ struct class_member *type__find_member_by_name(const struct type *type,
 uint32_t type__nr_members_of_type(const struct type *type, const type_id_t oftype);
 struct class_member *type__last_member(struct type *type);
 
+void enumeration__calc_prefix(struct type *type, const struct cu *cu);
+const char *enumeration__prefix(struct type *type, const struct cu *cu);
+uint16_t enumeration__prefix_len(struct type *type, const struct cu *cu);
+
+void enumerations__calc_prefix(struct list_head *enumerations);
+
 size_t typedef__fprintf(const struct tag *tag_type, const struct cu *cu,
 			const struct conf_fprintf *conf, FILE *fp);
 
@@ -1229,10 +1286,21 @@ static inline struct array_type *tag__array_type(const struct tag *tag)
 	return (struct array_type *)tag;
 }
 
+struct string_type {
+	struct tag      tag;
+	uint32_t        nr_entries;
+};
+
+static inline struct string_type *tag__string_type(const struct tag *tag)
+{
+	return (struct string_type *)tag;
+}
+
 struct enumerator {
 	struct tag	 tag;
 	strings_t	 name;
 	uint32_t	 value;
+	struct tag_cu	 type_enum; // To cache the type_enum searches
 };
 
 static inline const char *enumerator__name(const struct enumerator *enumerator,
@@ -1257,4 +1325,6 @@ void dwarves_print_version(FILE *fp, struct argp_state *state);
 
 extern bool no_bitfield_type_recode;
 
+extern const char tabs[];
+
 #endif /* _DWARVES_H_ */
