@@ -173,6 +173,8 @@ out:
 	return err;
 }
 
+bool ctf__verbose;
+
 struct ctf *ctf__new(const char *filename, Elf *elf)
 {
 	struct ctf *ctf = zalloc(sizeof(*ctf));
@@ -206,7 +208,8 @@ struct ctf *ctf__new(const char *filename, Elf *elf)
 		}
 
 		if (gelf_getehdr(ctf->elf, &ctf->ehdr) == NULL) {
-			fprintf(stderr, "%s: cannot get elf header.\n", __func__);
+			if (ctf__verbose)
+				fprintf(stderr, "%s: cannot get elf header.\n", __func__);
 			goto out_elf_end;
 		}
 
@@ -740,7 +743,7 @@ found_SUNW_ctf_str:
 	if (close(fd) < 0)
 		goto out_unlink;
 
-	char cmd[PATH_MAX];
+	char cmd[PATH_MAX * 2];
 	snprintf(cmd, sizeof(cmd), "objcopy --add-section .SUNW_ctf=%s %s",
 		 pathname, ctf->filename);
 	if (system(cmd) == 0)
